"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.createPlanton = void 0;
const delay_1 = __importDefault(require("delay"));
// eslint-disable-next-line @typescript-eslint/ban-ts-comment
// @ts-ignore
const promise_deferred_1 = __importDefault(require("promise-deferred"));
const serialize_error_1 = require("serialize-error");
const Logger_1 = __importDefault(require("../Logger"));
const errors_1 = require("../errors");
const createEmitter_1 = require("./createEmitter");
const log = Logger_1.default.child({
    namespace: 'createPlanton',
});
const createPlanton = (configuration) => {
    const { getActiveTaskInstructions, } = configuration;
    const events = createEmitter_1.createEmitter();
    const tasks = [];
    for (const inputTask of configuration.tasks) {
        log.debug('registered %s task', inputTask.name);
        for (const existingTask of tasks) {
            if (existingTask.name === inputTask.name) {
                throw new errors_1.DuplicateTaskNameError(existingTask.name);
            }
        }
        const calculateDelay = inputTask.delay || (() => {
            return 1000;
        });
        const concurrency = inputTask.concurrency === undefined ? 1 : inputTask.concurrency;
        if (concurrency < 1) {
            throw new errors_1.InvalidTaskConfigurationNameError(inputTask.name, 'Task concurrency must be greater than 0.');
        }
        const task = {
            attemptNumber: 0,
            concurrency,
            name: inputTask.name,
            schedule: inputTask.schedule,
        };
        const terminate = (() => {
            let delayPromise;
            const deferredTermination = new promise_deferred_1.default();
            let active = true;
            // eslint-disable-next-line complexity
            (async () => {
                // eslint-disable-next-line no-unmodified-loop-condition
                while (active) {
                    const calculatedDelay = calculateDelay(task.attemptNumber || 0);
                    if (calculatedDelay) {
                        delayPromise = delay_1.default(calculatedDelay);
                        await delayPromise;
                    }
                    if (!active) {
                        break;
                    }
                    const activeTaskInstructions = await getActiveTaskInstructions(inputTask.name);
                    if (activeTaskInstructions.length >= concurrency) {
                        continue;
                    }
                    let taskInstructions;
                    const limit = concurrency - activeTaskInstructions.length;
                    if (limit < 1) {
                        throw new errors_1.UnexpectedStateError('Limit cannot be less than 1.');
                    }
                    try {
                        taskInstructions = await inputTask.schedule({
                            activeTaskInstructions,
                            limit,
                        });
                    }
                    catch (error) {
                        log.error({
                            error: serialize_error_1.serializeError(error),
                            taskName: task.name,
                        }, 'scheduler produced an error');
                        events.emit('error', {
                            error,
                            taskName: task.name || '',
                        });
                        taskInstructions = [];
                    }
                    if (!Array.isArray(taskInstructions)) {
                        events.emit('error', {
                            error: new errors_1.UnexpectedTaskInstructionsError(task.name || '', taskInstructions),
                            taskName: task.name || '',
                        });
                        log.error({
                            taskInstructions,
                            taskName: task.name,
                        }, 'scheduler produced an unexpected result; result is not array');
                        taskInstructions = [];
                    }
                    if (taskInstructions.length > limit) {
                        events.emit('error', {
                            error: new errors_1.UnexpectedTaskInstructionsError(task.name || '', taskInstructions),
                            taskName: task.name || '',
                        });
                        log.error({
                            taskInstructions,
                            taskName: task.name,
                        }, 'scheduler produced an unexpected result; instruction number is greater than the limit');
                        taskInstructions = [];
                    }
                    for (const taskInstruction of taskInstructions) {
                        if (typeof taskInstruction !== 'string') {
                            events.emit('error', {
                                error: new errors_1.UnexpectedTaskInstructionsError(task.name || '', taskInstructions),
                                taskName: task.name || '',
                            });
                            log.error({
                                taskInstructions,
                                taskName: task.name,
                            }, 'scheduler produced an unexpected result; array members are not string');
                            taskInstructions = [];
                            break;
                        }
                    }
                    if (taskInstructions.length > 0) {
                        task.attemptNumber = 0;
                        for (const taskInstruction of taskInstructions) {
                            events.emit('task', {
                                instruction: taskInstruction,
                                taskName: task.name || '',
                            });
                        }
                    }
                    else if (task.attemptNumber !== undefined) {
                        task.attemptNumber++;
                    }
                    if (!active) {
                        break;
                    }
                }
                deferredTermination.resolve();
            })();
            return () => {
                active = false;
                if (delayPromise) {
                    delayPromise.clear();
                }
                return deferredTermination.promise;
            };
        })();
        // eslint-disable-next-line @typescript-eslint/ban-ts-comment
        // @ts-ignore
        task.terminate = terminate;
        tasks.push(task);
    }
    return {
        events,
        terminate: async () => {
            await Promise.all(tasks.map((task) => {
                return task.terminate();
            }));
        },
    };
};
exports.createPlanton = createPlanton;
